from tensorflow import keras
from pyhectiqlab import Run

class KerasCallback(keras.callbacks.Callback):
    
    def __init__(self, run: Run, 
                    level: str= "batch", 
                    exclude_train_metrics: bool = False,
                    exclude_val_metrics: bool = False,
                    exclude_predict_metrics: bool = False,
                    exclude_metrics=[]):
        """
        Keras callback to push the metrics in Hectiq Lab.

        level: "batch" or "epoch". Use epoch to track only the metics each epoch (the stepstamp will be the epoch number.)
        exclude_metrics: List the metrics to remove from tracking. For validation metrics, add "val_". For
            instance, use `exclude_metrics=["mse", "val_mse"]` to exclude the `mse` metrics in training and validation.

        Example usage:
        --------------
            model.fit(
                x_train,
                y_train,
                callbacks=[KerasCallback(run=run)],
            )
        """
        assert level in ["batch", "epoch"]
        self.level = level
        self.exclude_train_metrics = exclude_train_metrics
        self.exclude_val_metrics = exclude_val_metrics
        self.exclude_predict_metrics = exclude_predict_metrics
        self.exclude_metrics = exclude_metrics
        self._run = run
        self.epoch = 0
        self.global_step = 0
    
    def on_train_begin(self, logs=None):
        self._run.training()

    def on_train_end(self, logs=None):
        self._run.completed()

    def on_test_begin(self, logs=None):
        pass

    def on_test_end(self, logs=None):
        if self.level=="epoch":
            return

        if logs is None:
            return

        if self.exclude_val_metrics:
            return

        for metric in logs.keys():
            value = logs[metric]
            if metric in self.exclude_metrics:
                continue
            self._run.add_metrics(f"val_{metric}", step=self.global_step, value=value)

    def on_predict_begin(self, logs=None):
        pass

    def on_predict_end(self, logs=None):
        if logs is None:
            return

        if self.exclude_predict_metrics:
            return

        for metric in logs.keys():
            value = logs[metric]
            if metric in self.exclude_metrics:
                continue
                
            self._run.add_metrics(f"predict_{metric}", step=self.global_step, value=value)

    def on_train_batch_begin(self, batch, logs=None):
        pass

    def on_train_batch_end(self, batch, logs=None):
        self.global_step += 1
        
        if self.level=="epoch":
            return

        if logs == None:
            return

        if self.exclude_train_metrics:
            return

        for metric in logs.keys():
            value = logs[metric]
            if metric in self.exclude_metrics:
                continue
            self._run.add_metrics(f"train_{metric}", step=self.global_step, value=value)

    def on_test_batch_begin(self, batch, logs=None):
        pass

    def on_test_batch_end(self, batch, logs=None):
        pass

    def on_predict_batch_begin(self, batch, logs=None):
        pass

    def on_predict_batch_end(self, batch, logs=None):
        pass

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch = epoch

    def on_epoch_end(self, epoch, logs=None):

        if self.level=="batch":
            return

        for metric in logs.keys():
            value = logs[metric]
            if metric in self.exclude_metrics:
                continue
            if metric.startswith("val_"):
                if self.exclude_val_metrics:
                    continue
            else:
                if self.exclude_train_metrics:
                    continue
            self._run.add_metrics(metric, step=epoch, value=value)

        self.epoch = epoch+1




import os
from pyhectiqlab.utils import load_event_manager

import requests
from tqdm import tqdm
import logging
import json
import datetime as dt

logger = logging.getLogger('pyhectiqlab')

use_tqdm = os.getenv("HECTIQLAB_NO_TQDM", "0")!="1"

def app_info_at_path(dirpath: str):
    try:
        with open(os.path.join(dirpath, '.hlab_meta.json'), "r") as f:
            data = json.load(f)
        return data
    except:
        return None

def download_app(app_name: str,
                 project_path: str, 
                 save_path: str = './', 
                 overwrite: bool = False,
                 no_dir: bool = False):
    assert app_name, project_path
    manager = load_event_manager()
    if no_dir==False:
        dirpath = os.path.join(save_path, app_name)
        if os.path.isdir(dirpath) and overwrite==False:
            logger.warning(f'App is already downloaded at {dirpath}. Delete this folder or set overwrite==True to download again.')
            return dirpath
    else:
        dirpath = save_path

    if os.path.isdir(dirpath)==False:
        os.makedirs(dirpath)

    page = 1
    def fetch_download_urls(page: int):
        # Fetch the download files
        res = manager.add_event('get_app_download_info', 
                    args=(app_name, project_path, page), 
                    async_method=False)
        assert res.get('status_code')==200, res.get('detail')
        return res 

    res = fetch_download_urls(page)

    # Save the meta
    meta = res.get('meta')
    text = f"{meta.get('name')}\nProject: {project_path}\n"
    text += f"UUID: {meta.get('uuid')}\nDescription: {meta.get('short_description')}\n"
    text += f"Downloaded on: {dt.datetime.now()}"
    text += '\n--------------------------\n'

    with open(os.path.join(dirpath, 'HLab_README.md'), "w") as f:
        f.write(text)

    with open(os.path.join(dirpath, '.hlab_meta.json'), "w") as f:
        json.dump(meta, f)

    while True: 
        files = res.get('results', [])

        if len(files)==0:
            logger.info("Download completed.")
            return dirpath

        logger.info(f'Downloading {len(files)} files')
        for file_meta in files:
            url = file_meta['url']
            name = file_meta['name']
            if os.path.dirname(name)!='':
                subdirpath = os.path.join(dirpath, os.path.dirname(name))
                if os.path.isdir(subdirpath) == False:
                    os.makedirs(subdirpath)
            logger.info(name)
            response = requests.get(url, stream=True)
            total_size_in_bytes= int(response.headers.get('content-length', 0))
            block_size = 1024
            content_downloaded = 0
            if use_tqdm:
                progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)

            path = os.path.join(dirpath, name)
            with open(path, 'wb') as file:
                for data in response.iter_content(block_size):
                    content_downloaded += len(data)
                    if use_tqdm:
                        progress_bar.update(len(data))
                    file.write(data)
            if use_tqdm:
                progress_bar.close()

            if total_size_in_bytes != 0 and content_downloaded != total_size_in_bytes:
                logger.error(f"Something went wrong when downloading the file {path}.")
            logger.info(f'Content saved at {path}')

        # Next page
        page += 1
        res = fetch_download_urls(page)

    return dirpath





import os
import requests
import io
from tqdm import tqdm
from google.resumable_media.requests import ResumableUpload
import logging
logger = logging.getLogger('pyhectiqlab')

use_tqdm = os.getenv("HECTIQLAB_NO_TQDM", "0")!="1"
def upload_by_fragments(filepath: str, resumable_session_url: str, chunk_size: int = 1024*1024*4):
    """
    chunk_size: Size of the fragments (1024*1024==1Mb)
    """
    logger.debug(f'Fragment upload for {filepath}\nresumable_session_url: {resumable_session_url}\nchunk_size: {chunk_size}')

    upload = ResumableUpload(
        upload_url=resumable_session_url,
        chunk_size=chunk_size
    )
    data = open(filepath, 'rb').read()
    upload._stream = io.BytesIO(data)
    upload._total_bytes = len(data)
    upload._resumable_url = resumable_session_url
    
    transport = requests.Session()
    if use_tqdm:
        progress_bar = tqdm(total=upload.total_bytes, unit='iB', unit_scale=True)

    bytes_uploaded = 0
    while upload.finished==False:
        res = upload.transmit_next_chunk(transport)
        if use_tqdm:
            progress_bar.update(upload.bytes_uploaded-bytes_uploaded)
        bytes_uploaded = upload.bytes_uploaded
        assert res.status_code in [200,308], f'An error occured while uploading the file {filepath}\n{res.status_code}:: {res.text}'
    return 

def upload_with_signed_url(filepath: str, bucket_name: str, url: str, fields: dict):
    logger.debug(f'Single upload for {filepath}\nbucket_name: {bucket_name}\nurl: {url}')
    content_bytes = open(filepath, "rb")
    files = {"file": (bucket_name, content_bytes)}
    requests.post(url, data=fields, files=files)

def upload_file(filepath: str, policy: dict):
    """
    """
    if policy.get("active"):
        # Skip if file is active
        return
    upload_method = policy.get('upload_method')
    if upload_method=='fragment':
        logger.debug(f'Fragment upload for {filepath}')
        upload_by_fragments(filepath, 
            resumable_session_url=policy.get('url'))
    elif upload_method is None or upload_method=='single':
        logger.debug(f'Single upload for {filepath}')

        creds = policy.get('policy')
        upload_with_signed_url(filepath, 
            bucket_name=policy.get('bucket_name'),
            url=creds.get('url'),
            fields=creds.get('fields')
            )
    else:
        assert False, f'Unrecognized upload_method ({upload_method}).'
    logger.debug(f'Upload completed for {filepath}')

    return


import click
import getpass
import pyhectiqlab
import pyhectiqlab.ops as ops
from pyhectiqlab.auth import AuthProvider
from pyhectiqlab.mlmodels import download_mlmodel as ops_download_mlmodel
from pyhectiqlab.datasets import download_dataset as ops_download_dataset
from pyhectiqlab.apps import download_app as ops_download_app
import socket

@click.group()
def cli():
    """The official hectiqlab command line client"""
    pass

@cli.command()
def add_profile():
	auth = AuthProvider()
	username = input("Username: ")

	if auth.profile_exists(username):
		click.echo(f'A profile already exists for {username}')
		return

	password = getpass.getpass(prompt='Password: ', stream=None) 
	click.echo("Connecting...")
	success, api_key_uuid = auth.fetch_secret_api_token(username, password)

	if success:
		click.echo(f'Added profile [{username}] in {auth.tokens_path}')
		try:
			api_name = socket.gethostname()
			ops.update_secret_api_token_name(api_key_uuid, name=api_name, token=auth.secret_api_key)
			click.echo(f'Set the API-key name to {api_name}.')
		except:
			return
	else:
		click.echo('Unsuccessful login.')

@cli.command()
def version():
	click.echo(pyhectiqlab.__version__)

@cli.command()
def profiles():
	auth = AuthProvider()
	profiles = list(auth.profiles.keys())
	click.echo(profiles)

@cli.command()
@click.option('-p', '--project', prompt="Project", help='Path of the project (e.g. `hectiqai/project`)', required=True)
@click.option('-n', '--name', prompt='MLModel name', help='Name of the mlmodel', required=True)
@click.option('-v', '--version', help='Version of the mlmodel. If not specified, will download the latest release', required=False)
@click.option('-s', '--save_path', help='Save path [./]', default="./", required=False)
@click.option('-o', '--overwrite', is_flag=True)
def download_mlmodel(project, name, version, save_path, overwrite):
	dir_path = ops_download_mlmodel(mlmodel_name=name, 
						project_path=project, 
						version=version, 
						save_path=save_path, 
						overwrite=overwrite)
	if dir_path is not None:
		click.echo(f'MLModel saved in {dir_path}')

@cli.command()
@click.option('-p', '--project', prompt="Project", help='Path of the project (e.g. `hectiqai/project`)', required=True)
@click.option('-n', '--name', prompt='Dataset name', help='Name of the dataset', required=True)
@click.option('-v', '--version', help='Version of the dataset. If not specified, will download the latest release', required=False)
@click.option('-s', '--save_path', help='Save path [./]', default="./", required=True)
@click.option('-o', '--overwrite', is_flag=True)
def download_dataset(project, name, version, save_path, overwrite):
	dir_path = ops_download_dataset(dataset_name=name, 
						project_path=project, 
						version=version, 
						save_path=save_path, 
						overwrite=overwrite)
	if dir_path is not None:
		click.echo(f'Dataset saved in {dir_path}')

@cli.command()
@click.option('-p', '--project', prompt="Project", help='Path of the project (e.g. `hectiqai/project`)', required=True)
@click.option('-n', '--name', prompt='App name', help='Name of the app', required=True)
@click.option('-s', '--save_path', help='Save path [./]', default="./", required=True)
@click.option('-o', '--overwrite', is_flag=True)
@click.option('-r', '--raw', help='If set, the files are downloaded in save_path without directory', is_flag=True)
def download_app(project, name, save_path, overwrite, raw):
	dir_path = ops_download_app(app_name=name, 
						project_path=project, 
						save_path=save_path, 
						overwrite=overwrite,
						no_dir=raw)
	if dir_path is not None:
		click.echo(f'App saved in {dir_path}')
